#!/usr/bin/perl -w
##
# DCPU-16 assembler I/O functions for reading and writing address dumps.
#
# Address dumps consist of address prefixed word values, eg:
#   0000 : 1234
# or multiple word values:
#   0000 : 1234 5678 abce
#
# Address dumps contain no symbols or relocation data. They're just raw
# dumps of the content.
#
# @file
# @author Justin Fletcher
#

package DASMIO::Address;

use DASM;


##
# Create an object upon which we can work.
#
# @param[in] $dasm   DASM object we're working on
#
# @return new object
sub new
{
    my $proto = shift;
    my $dasm  = shift;
    my $class = ref($proto) || $proto;
    my $self  = {};

    bless $self, $class;
    
    $self->{'dasm'} = $dasm;
    
    return $self;
}


##
# Identify whether we can handle reading or writing a given file
# by its filename.
#
# @param[in]  $self     Address object
# @param[in]  $filename Filename we're accessing
# @param[in]  $type     'read' or 'write'
#
# @retval 1 if we can access it
# @retval 0 if we cannot access it
sub identify
{
    my ($self, $filename, $type) = @_;
    
    if ($filename =~ /\.dump$/)
    {
        return 1;
    }
    
    return 0;
}


##
# List the extensions we support
#
# @param[in]  $self     Address object
#
# @return arrayref of extensions
sub extensions
{
    return [ 'dump' ];
}


##
# Read an address dump and populate our core.
#
# @param[in]  $self     Address object
# @param[in]  $content  Content to parse
# @param[in]  $filename File it came from
sub read
{
    my ($self, $content, $filename) = @_;
    my $dasm = $self->{'dasm'};
    
    my $pc = $dasm->{'pc'};
    for my $line (split /\n/, $content)
    {
        # Step to the next line for now that we're parsing it
        $dasm->stepline($pc);

        if ($line =~ s/^\s*([0-9a-fA-F]+)\s*://)
        {
            # New base address at the front of the line
            $pc = hex($1);
        }
        my @vals = split /\s+/, $line;
        for my $val (@vals)
        {
            next if ($val eq '');
            
            # Check that we only allow hex numbers, possibly prefixed by
            # 0x.
            if ($val =~ /^(0x)?[0-9a-fA-F]+$/)
            {
                $dasm->store($pc++, hex($val));
            }
            else
            {
                $dasm->error("Cannot parse hex value '$val'");
            }
        }
    }
}


##
# Write an address dump from our core.
#
# In order to write an address dump, we need to force a resolve of all
# of the symbols. If there are any symbols that cannot be resolved, that's
# a fatal error - because they would write out as 0s.
#
# @param[in]  $self     Address object
#
# @return     Content to write
sub write
{
    my ($self) = @_;
    my $dasm = $self->{'dasm'};
    my $core = $dasm->{'core'};
    
    if ($dasm->resolve())
    {
        # everything was resolved
    }
    else
    {
        $dasm->error("Could not resolve all symbols to write out Address file");
    }
    
    my @content;
    my $row = "";
    my $lastaddress = 0;
    for my $address (sort { $a <=> $b } keys %$core)
    {
        # If we moved to a different address range (multiple of 8)
        # write out the row
        if ($row ne '' &&
            ($address & ~7) != ($lastaddress & ~7))
        {
            push @content, $row;
            $row = "";
        }

        # If this is the start of a line, put the address on the front
        if ($row eq '')
        {
            $row = sprintf "%04x :", $address;    
        }
        
        # Add on the word from the core
        $row .= sprintf " %04x", $core->{$address};
        
        # Remember where we were
        $lastaddress = $address;
    }
    if ($row ne "")
    {
        push @content, $row;
    }
    
    return join("\n", @content) . "\n";
}


# Return a true value
1;
